/////////////////////////////////////////////////////////////////////////
// adam setapen (asetapen)                  how to make (almost) anything
// 12.12.2010                                               MIT media lab
// bow_board.c
//      Capacitive sensor + accelerometer
//
//   0110000101110011011001010111010001100001011100000110010101101110
//   0110                       _                                0110
//   0001         __ _ ___  ___| |_ __ _ _ __   ___ _ __         0001
//   0110        / _` / __|/ _ \ __/ _` | '_ \ / _ \ '_ \        0110
//   1101       | (_| \__ \  __/ || (_| | |_) |  __/ | | |       1101
//   0111        \__,_|___/\___|\__\__,_| .__/ \___|_| |_|       0111
//   0011                               |_|                      0011
//   0110000101110011011001010111010001100001011100000110010101101110
/////////////////////////////////////////////////////////////////////////

// LED-switch program
#include <avr/io.h>
#include <util/delay.h>

#define TRUE 1
#define FALSE 0

// ONLY SET ONE OF THESE!
#define ENABLE_CAP_SENSOR 0
#define ENABLE_ACCEL 0
#define ENABLE_ALL 1

#define HIGH_SENSITIVITY 1

#define FILTER_1_CUTOFF 400
#define FILTER_2_CUTOFF 450
#define MAX_CAP_VAL 1023

#define output(directions,pin) (directions |= pin) // set port direction for output
#define set(port,pin) (port |= pin) // set port pin
#define clear(port,pin) (port &= (~pin)) // clear port pin
#define pin_test(pins,pin) (pins & pin) // test for port pin
#define bit_test(byte,bit) (byte & (1 << bit)) // test for bit set
#define bit_delay_time 100 // bit delay for 9600 with overhead
#define bit_delay() _delay_us(bit_delay_time) // RS232 bit delay
#define half_bit_delay() _delay_us(bit_delay_time/2) // RS232 half bit delay
#define charge_delay_1() _delay_us(1) // charge delay 1
#define charge_delay_2() _delay_us(10) // charge delay 2
#define charge_delay_3() _delay_us(100) // charge delay 3
#define settle_delay() _delay_us(100) // settle delay
#define char_delay() _delay_ms(10) // char delay

#define serial_port PORTB
#define serial_direction DDRB
#define serial_pin_out (1 << PB1)

#define charge_port PORTA
#define charge_direction DDRA
#define charge_pin (1 << PA7)
#define sense_port PORTA
#define sense_direction DDRA
#define sense_pin (1 << PA3)

#define accel_sleep_port PORTB
#define accel_sleep_direction DDRB
#define accel_sleep_pin (1 << PB2)

#define accel_select_port PORTA
#define accel_select_direction DDRA
#define accel_select_pin (1 << PA6)

void put_char(volatile unsigned char *port, unsigned char pin, char txchar) {
	// send character in txchar on port pin
	//    assumes line driver (inverts bits)
	//
	// start bit
	clear(*port,pin);
	bit_delay();
	
	// unrolled loop to write data bits
	//
	if bit_test(txchar,0)
	   set(*port,pin);
	else
	   clear(*port,pin);
	
	bit_delay();
	
	if bit_test(txchar,1)
	   set(*port,pin);
	else
	   clear(*port,pin);
	
	bit_delay();
	
	if bit_test(txchar,2)
	   set(*port,pin);
	else
	   clear(*port,pin);
	
	bit_delay();
	
	if bit_test(txchar,3)
	   set(*port,pin);
	else
	   clear(*port,pin);
	
	bit_delay();
	
	if bit_test(txchar,4)
	   set(*port,pin);
	else
	   clear(*port,pin);
	
	bit_delay();
	
	if bit_test(txchar,5)
	   set(*port,pin);
	else
	   clear(*port,pin);
	
	bit_delay();
	
	if bit_test(txchar,6)
	   set(*port,pin);
	else
	   clear(*port,pin);
	
	bit_delay();
	
	if bit_test(txchar,7)
	   set(*port,pin);
	else
	   clear(*port,pin);
	
	bit_delay();

	// stop bit
	set(*port,pin);
	bit_delay();

	// char delay
	bit_delay();
}

int is_touched(int *last1, int *last2) {	
	// Variables for capacitive info
	static unsigned char up_lo, up_hi, down_lo, down_hi;
	unsigned int uval1, dval1, uval2, dval2, val1, val2;
	unsigned int filter1, filter2;
	
	// Set ADC
	ADMUX = (0 << REFS1) | (0 << REFS0) // Vcc ref
	| (0 << MUX5) | (0 << MUX4) | (0 << MUX3) | (0 << MUX2) | (1 << MUX1) | (1 << MUX0); // PA3
	
	// settle, charge, and wait 1
	settle_delay();
	set(charge_port, charge_pin);
	charge_delay_1();
	
	// initiate conversion
	ADCSRA |= (1 << ADSC);
	
	// wait for completion
	while (ADCSRA & (1 << ADSC))
	   ;
	
	// save result
	up_lo = ADCL;
	up_hi = ADCH;
	
	// settle, discharge, and wait 1
	settle_delay();
	clear(charge_port, charge_pin);
	charge_delay_1();
	
	// initiate conversion
	ADCSRA |= (1 << ADSC);
	
	// wait for completion
	while (ADCSRA & (1 << ADSC))
	   ;
	
	// save result
	down_lo = ADCL;
	down_hi = ADCH;
	
	// Store result (val1)
	uval1 = 256 * up_hi + up_lo;
	dval1 = 256 * down_hi + down_lo;
	val1 = (uval1 + (MAX_CAP_VAL - dval1)) / 2;
	
	///////////////////////////////////////////////////////////////////////////
	// settle, charge, and wait 2
	settle_delay();
	set(charge_port, charge_pin);
	charge_delay_2();
	
	// initiate conversion
	ADCSRA |= (1 << ADSC);
	
	// wait for completion
	while (ADCSRA & (1 << ADSC))
	   ;
	
	// save result
	up_lo = ADCL;
	up_hi = ADCH;
	
	// settle, discharge, and wait 2
	settle_delay();
	clear(charge_port, charge_pin);
	charge_delay_2();
	
	// initiate conversion
	ADCSRA |= (1 << ADSC);
	
	// wait for completion
	while (ADCSRA & (1 << ADSC))
	   ;
	
	// save result
	down_lo = ADCL;
	down_hi = ADCH;
	
	// Store result (val2)
	uval2 = 256 * up_hi + up_lo;
	dval2 = 256 * down_hi + down_lo;
	val2 = (uval2 + (MAX_CAP_VAL - dval2)) / 2;
	
	// Filter data
	filter1 = ((*last1) / 2) + (val1 / 2);
	filter2 = ((*last2) / 2) + (val2 / 2);

	*last1 = filter1;
	*last2 = filter2;
	
	if (filter1 < FILTER_1_CUTOFF && filter2 < FILTER_2_CUTOFF) {
		return TRUE;
	} else {
		return FALSE;
	}
	
}

void init_board() {
	// set clock divider to /1
	CLKPR = (1 << CLKPCE);
	CLKPR = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0);
	
	// initialize output pins
	set(serial_port, serial_pin_out);
	output(serial_direction, serial_pin_out);
	clear(charge_port, charge_pin);
	output(charge_direction, charge_pin);
	set(accel_sleep_port, accel_sleep_pin);
	output(accel_sleep_direction, accel_sleep_pin);
	
	// Set sensitivity
	if (HIGH_SENSITIVITY) {
		set(accel_select_port, accel_select_pin);
		output(accel_select_direction, accel_select_pin);
	} else {
		clear(accel_select_port, accel_select_pin);
		output(accel_select_direction, accel_select_pin);
	}
	
	// init A/D
	ADMUX = (0 << REFS1) | (0 << REFS0) // Vcc ref
	| (0 << MUX5) | (0 << MUX4) | (0 << MUX3) | (0 << MUX2) | (1 << MUX1) | (1 << MUX0); // PA3
	ADCSRA = (1 << ADEN) // enable
	| (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); // prescaler /128
	
	// Set accel sleep mode off
	DDRB |= _BV(PB2);
}

void send_adc_data() {
	static unsigned char lo, hi;
	// initiate conversion
	ADCSRA |= (1 << ADSC);
	
	// wait for completion
	while (ADCSRA & (1 << ADSC))
	   ;
	
	// save result
	lo = ADCL;
	hi = ADCH;
	
	put_char(&serial_port, serial_pin_out, lo);
	char_delay();
	put_char(&serial_port, serial_pin_out, hi);
	char_delay();
}

//a function to write a DEC value to the serial port 
void put_decimal(int value){
	uint8_t temp[6];
	int8_t p = 0;
	uint8_t i;
	
	while(value > 0) {
		uint8_t cur_digit = value %10;
		temp[p++] = cur_digit + '0';
		value = value/10;
	}
	
	while(p>0){
		put_char(&serial_port, serial_pin_out,temp[--p]);
		char_delay();
	}
}


void send_accel_data() {
	// send framing
	put_char(&serial_port, serial_pin_out, 1);
	char_delay();
	put_char(&serial_port, serial_pin_out, 2);
	char_delay();
	put_char(&serial_port, serial_pin_out, 3);
	char_delay();
	put_char(&serial_port, serial_pin_out, 4);
	char_delay();
	
	// X Value
	ADMUX = (0 << REFS1) | (0 << REFS0) // Vcc ref
	| (0 << MUX5) | (0 << MUX4) | (0 << MUX3) | (0 << MUX2) | (0 << MUX1) | (0 << MUX0); // PA0
	
	send_adc_data();
	
	// Y Value
	ADMUX = (0 << REFS1) | (0 << REFS0) // Vcc ref
	| (0 << MUX5) | (0 << MUX4) | (0 << MUX3) | (0 << MUX2) | (0 << MUX1) | (1 << MUX0); // PA1
	
	send_adc_data();
	
	// Z Value
	ADMUX = (0 << REFS1) | (0 << REFS0) // Vcc ref
	| (0 << MUX5) | (0 << MUX4) | (0 << MUX3) | (0 << MUX2) | (1 << MUX1) | (0 << MUX0); // PA2
	
	send_adc_data();
}

int main() {
	int last1, last2;
	last1 = last2 = MAX_CAP_VAL;
	
	init_board();
	
	// main loop
	while (TRUE) {
		if (ENABLE_ALL) {
			if (is_touched(&last1, &last2)) {
				send_accel_data();
			}
		} else if (ENABLE_CAP_SENSOR) {
			if (is_touched(&last1, &last2)) {
				put_char(&serial_port, serial_pin_out, 'o');
				char_delay();
			}
		} else {
			send_accel_data();
		}
	}
}

